from scoop._control import _stat
import unittest

class TestStat(unittest.TestCase):
    def __init__(self, *args, **kwargs):
        super(TestStat, self).__init__(*args, **kwargs)

    def test_appendleft_with_zero(self):
        stats = _stat()
        data = list(range(0, 5))
        for i in data[::-1]:
            stats.appendleft(i)
        for i, j in enumerate(stats):
            self.assertEqual(data[i], j)

    def test_appendleft(self):
        stats = _stat()
        data = list(range(1, 5))
        for i in data[::-1]:
            stats.appendleft(i)
        for i, j in enumerate(stats):
            self.assertEqual(data[i], j)
     
    def test_maxlen(self):
        stats = _stat()
        data = list(range(1, 15))
        for i in data[::-1]:
            stats.appendleft(i)
        for i, j in enumerate(stats):
            self.assertEqual(data[i], j)
        self.assertEqual(len(stats), 10)

    def test_mean(self):
        stats = _stat()
        data = list(range(1, 11))
        for i in data:
            stats.appendleft(float(i))
        self.assertAlmostEqual(stats.mean(), 1.51044125730)
        stats.appendleft(1000)
        self.assertAlmostEqual(stats.mean(), 2.20121678520)

    def test_std(self):
        stats = _stat()
        data = list(range(1, 11))
        for i in data:
            stats.appendleft(float(i))
        self.assertAlmostEqual(stats.std(), 0.695407498476473)
        stats.appendleft(1000)
        self.assertAlmostEqual(stats.std(), 1.640541783885151)

    def test_mode(self):
        stats = _stat()
        data = list(range(1, 11))
        for i in data:
            stats.appendleft(float(i))
        self.assertAlmostEqual(stats.mode(), 2.79225543353)
        stats.appendleft(1000)
        self.assertAlmostEqual(stats.mode(), 0.61252803911)

    def test_median(self):
        stats = _stat()
        data = list(range(1, 11))
        for i in data:
            stats.appendleft(float(i))
        self.assertAlmostEqual(stats.median(), 4.52872868811)
        stats.appendleft(1000)
        self.assertAlmostEqual(stats.median(), 9.03600168611)

if __name__ == "__main__":
    t = unittest.TestLoader().loadTestsFromTestCase(TestStat)
    unittest.TextTestRunner(verbosity=2).run(t)



#
#    This file is part of Scalable COncurrent Operations in Python (SCOOP).
#
#    SCOOP is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation, either version 3 of
#    the License, or (at your option) any later version.
#
#    SCOOP is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with SCOOP. If not, see <http://www.gnu.org/licenses/>.
#
# Global imports
from collections import namedtuple
import logging
import os
import sys
import subprocess
from threading import Thread

# Local
import scoop
from scoop import utils
from .constants import BASE_SSH, BASE_RSH


class Host(object):
    """Represents an accessible computing resource.
       Can be remote (ssh via netowrk) or represent localhost."""
    LAUNCHING_ARGUMENTS = namedtuple(
        'launchingArguments',
        [
            'pythonPath', 'path', 'nice', 'pythonExecutable', 'size', 'origin',
            'brokerHostname', 'brokerPorts', 'debug', 'profiling', 'executable',
            'verbose', 'args', 'prolog', 'backend'
        ]
    )

    def __init__(self, hostname="localhost", rsh=False, ssh_executable='ssh'):
        self.workersArguments = None
        self.hostname = hostname
        self.subprocesses = []
        self.workerAmount = 0
        self.rsh = rsh
        self.ssh_executable = ssh_executable

    def __repr__(self):
        return "{0} ({1} workers)".format(
            self.hostname,
            self.workerAmount,
        )

    def isLocal(self):
        """Is the current host the localhost?"""
        return self.hostname in utils.localHostnames

    def setWorker(self, *args, **kwargs):
        """Add a worker assignation
            Arguments and order to pass are defined in LAUNCHING_ARGUMENTS
            Using named args is advised.
        """
        try:
            la = self.LAUNCHING_ARGUMENTS(*args, **kwargs)
        except TypeError as e:
            scoop.logger.error(("addWorker failed to convert args %s and kwargs %s "
                            "to namedtuple (requires %s arguments (names %s)") %
                            (args, kwargs, len(self.LAUNCHING_ARGUMENTS._fields),
                             self.LAUNCHING_ARGUMENTS._fields))

        self.workersArguments = la

    def setWorkerAmount(self, workerAmount):
        """Sets the worker amount to launch on this host."""
        self.workerAmount = workerAmount

    def _WorkerCommand_environment(self):
        """Return list of shell commands to prepare the environment for
           bootstrap."""
        worker = self.workersArguments
        c = []
        if worker.prolog:
            c.extend([
                "source",
                worker.prolog,
                "&&",
            ])
        if worker.pythonPath and not self.isLocal():
            # Tried to make it compliant to all shell variants.
            c.extend([
                "env",
                "PYTHONPATH={0}:$PYTHONPATH".format(worker.pythonPath),
            ])
        elif worker.pythonPath and self.isLocal():
            # Tried to make it compliant to all shell variants.
            c.extend([
                "env",
                "PYTHONPATH={0}:{1}".format(
                    worker.pythonPath,
                    os.environ.get("PYTHONPATH", ""),
                ),
            ])
        return c

    def _WorkerCommand_launcher(self):
        """Return list commands to start the bootstrap process"""
        return [
            self.workersArguments.pythonExecutable,
            '-m',
            'scoop.launch.__main__',
            str(self.workerAmount),
            str(self.workersArguments.verbose),
        ]

    def _WorkerCommand_options(self):
        """Return list of options for bootstrap"""
        worker = self.workersArguments
        c = []

        # If broker is on localhost
        if self.hostname == worker.brokerHostname:
            broker = "127.0.0.1"
        else:
            broker = worker.brokerHostname

        if worker.nice is not None:
            c.extend(['--nice', str(worker.nice)])
        c.extend(['--size', str(worker.size)])
        if self.isLocal():
            c.extend(['--workingDirectory', str(worker.path)])
        else:
            c.extend(['--workingDirectory', '"{0}"'.format(str(worker.path))])
        c.extend(['--brokerHostname', broker])
        c.extend(['--externalBrokerHostname', worker.brokerHostname])
        c.extend(['--taskPort', str(worker.brokerPorts[0])])
        c.extend(['--metaPort', str(worker.brokerPorts[1])])
        if worker.origin and worker.executable:
            c.append('--origin')
        if worker.debug:
            c.append('--debug')
        if worker.profiling:
            c.append('--profile')
        if worker.backend:
            c.append('--backend={0}'.format(worker.backend))
        if worker.verbose >= 1:
            c.append('-' + 'v' * worker.verbose)
        return c

    def _WorkerCommand_executable(self):
        """Return executable and any options to be executed by bootstrap"""
        worker = self.workersArguments

        c = []
        if worker.executable:
            c.append(worker.executable)
        # This trick is used to parse correctly quotes
        # (ie. myScript.py 'arg1 "arg2" arg3')
        # Because it is parsed by a shell, every quote gets re-interpreted.
        # It replaces simple quotation marks with \\\" which gets evaluated to
        # \" by the second shell which prints it out as a double quote.
        if worker.args:
            if self.isLocal():
                # If local, no shell is used so no escaping needed
                c.extend([
                    '{0}'.format(a) for a in worker.args
                ])
            else:
                c.extend([
                    '"{0}"'.format(a.replace('"', '\\\"')) for a in worker.args
                ])
        return c

    def _getWorkerCommandList(self):
        """Generate the workerCommand as list"""
        c = []
        c.extend(self._WorkerCommand_environment())
        c.extend(self._WorkerCommand_launcher())
        c.extend(self._WorkerCommand_options())
        c.extend(self._WorkerCommand_executable())

        return c

    def getCommand(self):
        """Retrieves the shell command to launch the workers on this host."""
        return " ".join(self._getWorkerCommandList())

    def launch(self, tunnelPorts=None):
        """Launch every worker assigned on this host."""
        if self.isLocal():
            # Launching local workers
            c = self._getWorkerCommandList()
            self.subprocesses.append(subprocess.Popen(c))
        else:
            # Launching remotely
            BASE_SSH[0] = self.ssh_executable
            sshCmd = BASE_SSH if not self.rsh else BASE_RSH
            if tunnelPorts is not None:
                sshCmd += [
                    '-R {0}:127.0.0.1:{0}'.format(tunnelPorts[0]),
                    '-R {0}:127.0.0.1:{0}'.format(tunnelPorts[1]),
                ]
            self.subprocesses.append(
                subprocess.Popen(sshCmd + [self.hostname, self.getCommand()],
                                 bufsize=-1,
                                 stdout=None,
                                 stderr=None,
                                 stdin=subprocess.PIPE
                )
            )

        return self.subprocesses

    def close(self):
        """Connection(s) cleanup."""
        # Ensure everything is cleaned up on exit
        scoop.logger.debug('Closing workers on {0}.'.format(self))

        # Terminate subprocesses
        for process in self.subprocesses:
            try:
                process.terminate()
                process.wait()
            except OSError:
                pass



#
#    This file is part of Scalable COncurrent Operations in Python (SCOOP).
#
#    SCOOP is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation, either version 3 of
#    the License, or (at your option) any later version.
#
#    SCOOP is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with SCOOP. If not, see <http://www.gnu.org/licenses/>.
#

# -tt : Enforce a shell to be created to hold the remote comand.
#      This allows the kill of the remote command on ssh's termination
# -x : Disable X11 redirection.
# -n : Attach stdin to /dev/null .
# -oStrictHostKeyChecking=no : Automatically accept host fingerprint.
# -oBatchMode=yes : Add automatically the host to known hosts.
# -oServerAliveInterval=300 : Send sign of life every 5 minutes to prevent
#                             the server to close the connection.
BASE_SSH = [
    '{{ssh_executable}}',
    '-tt',
    '-x',
    '-oStrictHostKeyChecking=no',
    '-oBatchMode=yes',
    '-oUserKnownHostsFile=/dev/null',
    '-oServerAliveInterval=300',
]

BASE_RSH = [
    'rsh',
]



#!/usr/bin/env python
#
#    This file is part of Scalable COncurrent Operations in Python (SCOOP).
#
#    SCOOP is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation, either version 3 of
#    the License, or (at your option) any later version.
#
#    SCOOP is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with SCOOP. If not, see <http://www.gnu.org/licenses/>.
#
import unittest
import os
from itertools import repeat

from scoop import utils

# This is a subset of a SGE environment for testing
SGE_ENV = {'NHOSTS':'2', 'PE_HOSTFILE':"sgehostssim.txt",
'NSLOTS': '16', 'PE': 'default', 'ENVIRONMENT': 'BATCH'}

# This is a subset of a PBS environment for testing
PBS_ENV = {'ENVIRONMENT': 'BATCH', 'PBS_ENVIRONMENT': 'PBS_BATCH',
'MOAB_PROCCOUNT': '16', 'PBS_NUM_NODES': '2', 'PBS_NP': '16', 
'PBS_NODEFILE': 'pbshostssim.txt'} 

# This is the logical content of the hostfiles
hosts = [("host1", 8), ("host2", 4), ("host3", 2), ("host4", 2)]

class TestUtils(unittest.TestCase):
    def __init__(self, *args, **kwargs):
        super(TestUtils, self).__init__(*args, **kwargs)

    def setUp(self):
        self.backup = os.environ.copy()

    def tearDown(self):
        os.environ = self.backup

    def test_getCpuCount(self):
        """Since we don't know how many cores has the computer on which these
        tests will be run, we check that at least, getCPUcount() returns an
        int larger than 0."""
        self.assertIsInstance(utils.getCPUcount(), int)
        self.assertTrue(utils.getCPUcount() > 0)

    def test_getEnvPBS(self):
        os.environ.update(PBS_ENV)
        self.assertEqual(utils.getEnv(), "PBS")

    def test_getEnvSGE(self):
        os.environ.update(SGE_ENV)
        self.assertEqual(utils.getEnv(), "SGE")

    def test_getEnvOther(self):
        self.assertEqual(utils.getEnv(), "other")

    def test_getHostsPBS(self):
        os.environ.update(PBS_ENV)
        # We used the set because the order of the hosts is not important.
        self.assertEqual(set(utils.getHosts()), set(hosts))

    def test_getHostsSGE(self):
        os.environ.update(SGE_ENV)
        # We used the set because the order of the hosts is not important.
        self.assertEqual(set(utils.getHosts()), set(hosts))

    def test_getHostsFromSLURM(self):
        pass

    def test_parseSLURM_dashOneDecimal(self):
        hosts = utils.parseSLURM("n[1-4]")
        result = zip(("n{0}".format(x) for x in range(1, 5)), repeat(1))
        self.assertEqual(set(hosts), set(result))

    def test_parseSLURM_dashTwoDecimals(self):
        hosts = utils.parseSLURM("n[5-10]")
        result = zip(("n{0:02d}".format(x) for x in range(5, 11)), repeat(1))
        self.assertEqual(set(hosts), set(result))

    def test_parseSLURM_dashTwonames(self):
        hosts = utils.parseSLURM("x[1-2]y[1-2]")
        result = []
        for num in range(1, 3):
            for prefix in ["x", "y"]:
                result.append(("{prefix}{num}".format(**locals()), 1))
        self.assertEqual(set(hosts), set(result))

    def test_parseSLURM_nondashOneDecimal(self):
        hosts = utils.parseSLURM("n[1,4]")
        result = []
        result = zip(("n{0}".format(x) for x in [1, 4]), repeat(1))
        self.assertEqual(set(hosts), set(result))

    def test_parseSLURM_nondash_and_dashOneDecimal(self):
        hosts = utils.parseSLURM("n[1,5-9]")
        result = []
        result = zip(("n{0}".format(x) for x in [1, 5, 6, 7, 8, 9]), repeat(1))
        self.assertEqual(set(hosts), set(result))
        
    def test_getHostsFile(self):
        self.assertEqual(set(utils.getHosts("hostfilesim.txt")), set(hosts))

    def test_getWorkerQtePBS(self):
        os.environ.update(PBS_ENV)
        self.assertEqual(utils.getWorkerQte(utils.getHosts()), 16)

    def test_getWorkerQteSGE(self):
        os.environ.update(SGE_ENV)
        self.assertEqual(utils.getWorkerQte(utils.getHosts()), 16)

    def test_getWorkerQteFile(self):
        self.assertEqual(utils.getWorkerQte(utils.getHosts("hostfilesim.txt")), 16)


if __name__ == "__main__":
    t = unittest.TestLoader().loadTestsFromTestCase(TestUtils)
    unittest.TextTestRunner(verbosity=2).run(t)



#
#    This file is part of Scalable COncurrent Operations in Python (SCOOP).
#
#    SCOOP is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation, either version 3 of
#    the License, or (at your option) any later version.
#
#    SCOOP is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with SCOOP. If not, see <http://www.gnu.org/licenses/>.
#
import time
import sys
import random
import socket
import copy
import logging
import threading
try:
    import cPickle as pickle
except ImportError:
    import pickle

import zmq

import scoop
from .. import shared, encapsulation, utils
from ..shared import SharedElementEncapsulation
from .scoopexceptions import Shutdown, ReferenceBroken

# Worker requests
INIT = b"I"
REQUEST = b"RQ"
TASK = b"T"
REPLY = b"RP"
SHUTDOWN = b"S"
VARIABLE = b"V"
BROKER_INFO = b"B"
STATUS_REQ = b"SR"
STATUS_ANS = b"SA"
STATUS_DONE = b"SD"
STATUS_UPDATE = b"SU"

# Task statuses
STATUS_HERE = b"H"
STATUS_GIVEN = b"G"
STATUS_NONE = b"N"


LINGER_TIME = 1000


class ZMQCommunicator(object):
    """This class encapsulates the communication features toward the broker."""

    def __init__(self):
        self.ZMQcontext = zmq.Context()

        # TODO number of broker
        self.number_of_broker = float('inf')
        self.broker_set = set()


        # Get the current address of the interface facing the broker
        info = socket.getaddrinfo(scoop.BROKER.externalHostname, scoop.BROKER.task_port)[0]
        s = socket.socket(info[0], socket.SOCK_DGRAM)
        s.connect(info[4][:2])
        external_addr = s.getsockname()[0]
        s.close()

        if external_addr in utils.loopbackReferences or info[0] == socket.AF_INET6:
            external_addr = scoop.BROKER.externalHostname

        # Create an inter-worker socket
        self.direct_socket_peers = []
        self.direct_socket = self.createZMQSocket(zmq.ROUTER)
        # TODO: This doesn't seems to be respected in the ROUTER socket
        self.direct_socket.setsockopt(zmq.SNDTIMEO, 0)
        # Code stolen from pyzmq's bind_to_random_port() from sugar/socket.py
        for i in range(100):
            try:
                self.direct_socket_port = random.randrange(49152, 65536)
                # Set current worker inter-worker socket name to its addr:port
                scoop.worker = "{addr}:{port}".format(
                    addr=external_addr,
                    port=self.direct_socket_port,
                ).encode()
                self.direct_socket.setsockopt(zmq.IDENTITY, scoop.worker)
                self.direct_socket.bind("tcp://*:{0}".format(
                    self.direct_socket_port,
                ))
            except:
                # Except on ZMQError with a check on EADDRINUSE should go here
                # but its definition is not consistent in pyzmq over multiple
                # versions
                pass
            else:
                break
        else:
            raise Exception("Could not create direct connection socket")

        # Update the logger to display our name
        try:
            scoop.logger.handlers[0].setFormatter(
                logging.Formatter(
                    "[%(asctime)-15s] %(module)-9s ({0}) %(levelname)-7s "
                    "%(message)s".format(scoop.worker)
                )
            )
        except IndexError:
            scoop.logger.debug(
                "Could not set worker name into logger ({0})".format(
                    scoop.worker
                )
            )

        # socket for the futures, replies and request
        self.socket = self.createZMQSocket(zmq.DEALER)
        self.socket.setsockopt(zmq.IDENTITY, scoop.worker)

        # socket for the shutdown signal
        self.infoSocket = self.createZMQSocket(zmq.SUB)

        # Set poller
        self.poller = zmq.Poller()
        self.poller.register(self.socket, zmq.POLLIN)
        self.poller.register(self.direct_socket, zmq.POLLIN)
        self.poller.register(self.infoSocket, zmq.POLLIN)

        self._addBroker(scoop.BROKER)

        # Send an INIT to get all previously set variables and share
        # current configuration to broker
        self.socket.send_multipart([
            INIT,
            pickle.dumps(scoop.CONFIGURATION)
        ])
        scoop.CONFIGURATION.update(pickle.loads(self.socket.recv()))
        inboundVariables = pickle.loads(self.socket.recv())
        shared.elements = dict([
            (pickle.loads(key),
                dict([(pickle.loads(varName),
                       pickle.loads(varValue))
                    for varName, varValue in value.items()
                ]))
                for key, value in inboundVariables.items()
        ])
        for broker in pickle.loads(self.socket.recv()):
            # Skip already connected brokers
            if broker in self.broker_set:
                continue
            self._addBroker(broker)

        # Putting futures status reporting in place
        self.status_update_thread = threading.Thread(target=self._reportFutures)
        self.status_update_thread.daemon = True
        self.status_update_thread.start()

    def createZMQSocket(self, sock_type):
        """Create a socket of the given sock_type and deactivate message dropping"""
        sock = self.ZMQcontext.socket(sock_type)
        sock.setsockopt(zmq.LINGER, LINGER_TIME)
        sock.setsockopt(zmq.IPV4ONLY, 0)

        # Remove message dropping
        sock.setsockopt(zmq.SNDHWM, 0)
        sock.setsockopt(zmq.RCVHWM, 0)
        try:
            sock.setsockopt(zmq.IMMEDIATE, 1)
        except:
            # This parameter was recently added by new libzmq versions
            pass

        # Don't accept unroutable messages
        if sock_type == zmq.ROUTER:
            sock.setsockopt(zmq.ROUTER_MANDATORY, 1)
        return sock

    def _reportFutures(self):
        """Sends futures status updates to broker at intervals of
        scoop.TIME_BETWEEN_STATUS_REPORTS seconds. Is intended to be run by a
        separate thread."""
        try:
            while True:
                time.sleep(scoop.TIME_BETWEEN_STATUS_REPORTS)
                fids = set(x.id for x in scoop._control.execQueue.movable)
                fids.update(set(x.id for x in scoop._control.execQueue.ready))
                fids.update(set(x.id for x in scoop._control.execQueue.inprogress))
                self.socket.send_multipart([
                    STATUS_UPDATE,
                    pickle.dumps(fids),
                ])
        except AttributeError:
            # The process is being shut down.
            pass

    def addPeer(self, peer):
        if peer not in self.direct_socket_peers:
            self.direct_socket_peers.append(peer)
            new_peer = "tcp://{0}".format(peer.decode("utf-8"))
            self.direct_socket.connect(new_peer)

    def _addBroker(self, brokerEntry):
        # Add a broker to the socket and the infosocket.
        broker_address = "tcp://{hostname}:{port}".format(
            hostname=brokerEntry.hostname,
            port=brokerEntry.task_port,
        )
        meta_address = "tcp://{hostname}:{port}".format(
            hostname=brokerEntry.hostname,
            port=brokerEntry.info_port,
        )
        self.socket.connect(broker_address)

        self.infoSocket.connect(meta_address)
        self.infoSocket.setsockopt(zmq.SUBSCRIBE, b"")

        self.broker_set.add(brokerEntry)

    def _poll(self, timeout):
        self.pumpInfoSocket()
        return self.poller.poll(timeout)

    def _recv(self):
        # Prioritize answers over new tasks
        if self.direct_socket.poll(0):
            router_msg = self.direct_socket.recv_multipart()
            # Remove the sender address
            msg = router_msg[1:] + [router_msg[0]]
        else:
            msg = self.socket.recv_multipart()
        
        try:
            thisFuture = pickle.loads(msg[1])
        except (AttributeError, ImportError) as e:
            scoop.logger.error(
                "An instance could not find its base reference on a worker. "
                "Ensure that your objects have their definition available in "
                "the root scope of your program.\n{error}".format(
                    error=e,
                )
            )
            raise ReferenceBroken(e)

        if msg[0] == TASK:
            # Try to connect directly to this worker to send the result
            # afterwards if Future is from a map.
            if thisFuture.sendResultBack:
                self.addPeer(thisFuture.id[0])

        elif msg[0] == STATUS_ANS:
            # TODO: This should not be here but in FuturesQueue.
            if msg[2] == STATUS_HERE:
                # TODO: Don't know why should that be done?
                self.sendRequest()
            elif msg[2] == STATUS_NONE:
                # If a task was requested but is nowhere to be found, resend it
                future_id = pickle.loads(msg[1])
                try:
                    scoop.logger.warning(
                        "Lost track of future {0}. Resending it..."
                        "".format(scoop._control.futureDict[future_id])
                    )
                    self.sendFuture(scoop._control.futureDict[future_id])
                except KeyError:
                    # Future was received and processed meanwhile
                    pass
            return

        isCallable = callable(thisFuture.callable)
        isDone = thisFuture._ended()
        if not isCallable and not isDone:
            # TODO: Also check in root module globals for fully qualified name
            try:
                module_found = hasattr(sys.modules["__main__"],
                                       thisFuture.callable)
            except TypeError:
                module_found = False
            if module_found:
                thisFuture.callable = getattr(sys.modules["__main__"],
                                              thisFuture.callable)
            else:
                raise ReferenceBroken("This element could not be pickled: "
                                      "{0}.".format(thisFuture))
        return thisFuture

    def pumpInfoSocket(self):
        try:
            while self.infoSocket.poll(0):
                msg = self.infoSocket.recv_multipart()
                if msg[0] == SHUTDOWN:
                    if scoop.IS_ORIGIN is False:
                        raise Shutdown("Shutdown received")
                    if not scoop.SHUTDOWN_REQUESTED:
                        scoop.logger.error(
                            "A worker exited unexpectedly. Read the worker logs "
                            "for more information. SCOOP pool will now shutdown."
                        )
                        raise Shutdown("Unexpected shutdown received")
                elif msg[0] == VARIABLE:
                    key = pickle.loads(msg[3])
                    varValue = pickle.loads(msg[2])
                    varName = pickle.loads(msg[1])
                    shared.elements.setdefault(key, {}).update({varName: varValue})
                    self.convertVariable(key, varName, varValue)
                elif msg[0] == BROKER_INFO:
                    # TODO: find out what to do here ...
                    if len(self.broker_set) == 0: # The first update
                        self.broker_set.add(pickle.loads(msg[1]))
                    if len(self.broker_set) < self.number_of_broker:
                        brokers = pickle.loads(msg[2])
                        needed = self.number_of_broker - len(self.broker_set)
                        try:
                            new_brokers = random.sample(brokers, needed)
                        except ValueError:
                            new_brokers = brokers
                            self.number_of_broker = len(self.broker_set) + len(new_brokers)
                            scoop.logger.warning(("The number of brokers could not be set"
                                            " on worker {0}. A total of {1} worker(s)"
                                            " were set.".format(scoop.worker,
                                                                self.number_of_broker)))

                        for broker in new_brokers:
                            broker_address = "tcp://" + broker.hostname + broker.task_port
                            meta_address = "tcp://" + broker.hostname + broker.info_port
                            self._addBroker(broker_address, meta_address)
                        self.broker_set.update(new_brokers)
        except zmq.error.ZMQError:
            pass

    def convertVariable(self, key, varName, varValue):
        """Puts the function in the globals() of the main module."""
        if isinstance(varValue, encapsulation.FunctionEncapsulation):
            result = varValue.getFunction()

            # Update the global scope of the function to match the current module
            mainModule = sys.modules["__main__"]
            result.__name__ = varName
            result.__globals__.update(mainModule.__dict__)
            setattr(mainModule, varName, result)
            shared.elements[key].update({
                varName: result,
            })

    def recvFuture(self):
        while self._poll(0):
            received = self._recv()
            if received:
                yield received

    def sendFuture(self, future):
        """Send a Future to be executed remotely."""
        future = copy.copy(future)
        future.greenlet = None
        future.children = {}

        try:
            if shared.getConst(hash(future.callable), timeout=0):
                # Enforce name reference passing if already shared
                future.callable = SharedElementEncapsulation(hash(future.callable))
            self.socket.send_multipart([
                TASK,
                pickle.dumps(future.id, pickle.HIGHEST_PROTOCOL),
                pickle.dumps(future, pickle.HIGHEST_PROTOCOL),
            ])
        except (pickle.PicklingError, TypeError) as e:
            # If element not picklable, pickle its name
            # TODO: use its fully qualified name
            scoop.logger.warn("Pickling Error: {0}".format(e))
            future.callable = hash(future.callable)
            self.socket.send_multipart([
                TASK,
                pickle.dumps(future.id, pickle.HIGHEST_PROTOCOL),
                pickle.dumps(future, pickle.HIGHEST_PROTOCOL),
            ])

    def sendResult(self, future):
        """Send a terminated future back to its parent."""
        future = copy.copy(future)

        # Remove the (now) extraneous elements from future class
        future.callable = future.args = future.kargs = future.greenlet = None

        if not future.sendResultBack:
            # Don't reply back the result if it isn't asked
            future.resultValue = None

        self._sendReply(
            future.id[0],
            pickle.dumps(future.id, pickle.HIGHEST_PROTOCOL),
            pickle.dumps(future, pickle.HIGHEST_PROTOCOL),
        )

    def _sendReply(self, destination, fid, *args):
        """Send a REPLY directly to its destination. If it doesn't work, launch
        it back to the broker."""
        # Try to send the result directly to its parent
        self.addPeer(destination)

        try:
            self.direct_socket.send_multipart([
                destination,
                REPLY,
            ] + list(args),
                flags=zmq.NOBLOCK)
        except zmq.error.ZMQError as e:
            # Fallback on Broker routing if no direct connection possible
            scoop.logger.debug(
                "{0}: Could not send result directly to peer {1}, routing through "
                "broker.".format(scoop.worker, destination)
            )
            self.socket.send_multipart([
                REPLY, 
                ] + list(args) + [
                destination,
            ])

        self.socket.send_multipart([
            STATUS_DONE,
            fid,
        ])

    def sendStatusRequest(self, future):
        self.socket.send_multipart([
            STATUS_REQ,
            pickle.dumps(future.id, pickle.HIGHEST_PROTOCOL),
        ])

    def sendVariable(self, key, value):
        self.socket.send_multipart([
            VARIABLE,
            pickle.dumps(key, pickle.HIGHEST_PROTOCOL),
            pickle.dumps(value, pickle.HIGHEST_PROTOCOL),
            pickle.dumps(scoop.worker, pickle.HIGHEST_PROTOCOL),
        ])

    def sendRequest(self):
        for _ in range(len(self.broker_set)):
            self.socket.send(REQUEST)

    def workerDown(self):
        self.socket.send(WORKERDOWN)

    def shutdown(self):
        """Sends a shutdown message to other workers."""
        if self.ZMQcontext and not self.ZMQcontext.closed:
            scoop.SHUTDOWN_REQUESTED = True
            self.socket.send(SHUTDOWN)
            
            # pyzmq would issue an 'no module named zmqerror' on windows
            # without this
            try:
                self.direct_socket.__del__()
                self.socket.__del__()
                self.infoSocket.__del__()
            except AttributeError:
                # Older versions of pyzmq doesn't have the __del__ method
                pass

            self.ZMQcontext.destroy()



#!/usr/bin/env python
#
#    This file is part of Scalable COncurrent Operations in Python (SCOOP).
#
#    SCOOP is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation, either version 3 of
#    the License, or (at your option) any later version.
#
#    SCOOP is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with SCOOP. If not, see <http://www.gnu.org/licenses/>.
#
"""This module is executed from the launcher via SSH and is used to launch the
workers using the scoop.bootstrap module. It can detect the number of cores on
the machine and simplify the SSH command.

Usage:
python -m scoop.launch [nb_to_launch] [arguments to the bootstrap module]"""

import sys
import os
from subprocess import Popen

from scoop.utils import getCPUcount

import atexit


BOOTSTRAP_MODULE = 'scoop.bootstrap.__main__'


def getArgs():
    """Gets the arguments of the program.
    Returns a tuple containting:
    (qty to launch, arguments to pass to the bootstrap module)."""
    try:
        nb_to_launch = int(sys.argv[1])
    except:
        nb_to_launch = 0

    if nb_to_launch == 0:
        nb_to_launch = getCPUcount()

    try:
        verbosity = int(sys.argv[2])
    except:
        verbosity = 3

    return nb_to_launch, verbosity, sys.argv[3:]


def cleanupBootstraps():
    """Perform a cleanup (terminate) of the children processes."""
    for p in processes:
        try:
            p.terminate()
        except OSError:
            pass


def launchBootstraps():
    """Launch the bootstrap instances in separate subprocesses"""
    global processes
    worker_amount, verbosity, args = getArgs()
    was_origin = False

    if verbosity >= 1:
        sys.stderr.write("Launching {0} worker(s) using {1}.\n".format(
                worker_amount,
                os.environ['SHELL'] if 'SHELL' in os.environ else "an unknown shell",
            )
        )
        sys.stderr.flush()

    processes = []
    for _ in range(worker_amount):
        command = [sys.executable, "-m", BOOTSTRAP_MODULE] + args
        if verbosity >= 3:
            sys.stderr.write("Executing '{0}'...\n".format(command))
            sys.stderr.flush()
        processes.append(Popen(command))

        # Only have a single origin
        try:
            args.remove("--origin")
        except ValueError:
            pass
        else:
            was_origin = True

    if was_origin:
        # Only wait on the origin, this will return and notify the launcher
        # the the job has finished and start the cleanup phase
        try:
            processes[0].wait()
        except KeyboardInterrupt:
            pass
    else:
        for p in processes:
            p.wait()


if __name__ == "__main__":
    processes = []
    try:
        launchBootstraps()
    finally:
        cleanupBootstraps()


#
#    This file is part of Scalable COncurrent Operations in Python (SCOOP).
#
#    SCOOP is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation, either version 3 of
#    the License, or (at your option) any later version.
#
#    SCOOP is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with SCOOP. If not, see <http://www.gnu.org/licenses/>.
#
"""
Code from the deap framework, available at:
https://code.google.com/p/deap/source/browse/examples/ga/evosn.py
Conversion to its parallel form took two lines:
from scoop import futures
toolbox.register("map", futures.map)
"""
import random

import numpy

from deap import algorithms
from deap import base
from deap import creator
from deap import tools
from scoop import futures

from dependency import sortingnetwork as sn

INPUTS = 6

def evalEvoSN(individual, dimension):
    network = sn.SortingNetwork(dimension, individual)
    return network.assess(), network.length, network.depth

def genWire(dimension):
    return (random.randrange(dimension), random.randrange(dimension))
    
def genNetwork(dimension, min_size, max_size):
    size = random.randint(min_size, max_size)
    return [genWire(dimension) for i in range(size)]
    
def mutWire(individual, dimension, indpb):
    for index, elem in enumerate(individual):
        if random.random() < indpb:
            individual[index] = genWire(dimension)      

def mutAddWire(individual, dimension):
    index = random.randint(0, len(individual))
    individual.insert(index, genWire(dimension))

def mutDelWire(individual):
    index = random.randrange(len(individual))
    del individual[index]

creator.create("FitnessMin", base.Fitness, weights=(-1.0, -1.0, -1.0))
creator.create("Individual", list, fitness=creator.FitnessMin)

toolbox = base.Toolbox()

# Gene initializer
toolbox.register("network", genNetwork, dimension=INPUTS, min_size=9, max_size=12)

# Structure initializers
toolbox.register("individual", tools.initIterate, creator.Individual, toolbox.network)
toolbox.register("population", tools.initRepeat, list, toolbox.individual)

toolbox.register("evaluate", evalEvoSN, dimension=INPUTS)
toolbox.register("mate", tools.cxTwoPoint)
toolbox.register("mutate", mutWire, dimension=INPUTS, indpb=0.05)
toolbox.register("addwire", mutAddWire, dimension=INPUTS)
toolbox.register("delwire", mutDelWire)
toolbox.register("select", tools.selNSGA2)

# Enable parallel computing using SCOOP
toolbox.register("map", futures.map)

def main():
    random.seed(64)

    population = toolbox.population(n=300)
    hof = tools.ParetoFront()
    
    stats = tools.Statistics(lambda ind: ind.fitness.values)
    stats.register("avg", numpy.mean, axis=0)
    stats.register("std", numpy.std, axis=0)
    stats.register("min", numpy.min, axis=0)
    stats.register("max", numpy.max, axis=0)
    
    logbook = tools.Logbook()
    logbook.header = "gen", "evals", "std", "min", "avg", "max"
    
    CXPB, MUTPB, ADDPB, DELPB, NGEN = 0.5, 0.2, 0.01, 0.01, 40
    
    # Evaluate every individuals
    fitnesses = toolbox.map(toolbox.evaluate, population)
    for ind, fit in zip(population, fitnesses):
        ind.fitness.values = fit
    
    hof.update(population)
    record = stats.compile(population)
    logbook.record(gen=0, evals=len(population), **record)
    print(logbook.stream)
    
    # Begin the evolution
    for g in range(1, NGEN):
        offspring = [toolbox.clone(ind) for ind in population]
    
        # Apply crossover and mutation
        for ind1, ind2 in zip(offspring[::2], offspring[1::2]):
            if random.random() < CXPB:
                toolbox.mate(ind1, ind2)
                del ind1.fitness.values
                del ind2.fitness.values
        
        # Note here that we have a different sheme of mutation than in the
        # original algorithm, we use 3 different mutations subsequently.
        for ind in offspring:
            if random.random() < MUTPB:
                toolbox.mutate(ind)
                del ind.fitness.values
            if random.random() < ADDPB:
                toolbox.addwire(ind)
                del ind.fitness.values
            if random.random() < DELPB:
                toolbox.delwire(ind)
                del ind.fitness.values
                
        # Evaluate the individuals with an invalid fitness
        invalid_ind = [ind for ind in offspring if not ind.fitness.valid]
        fitnesses = toolbox.map(toolbox.evaluate, invalid_ind)
        for ind, fit in zip(invalid_ind, fitnesses):
            ind.fitness.values = fit
        
        population = toolbox.select(population+offspring, len(offspring))
        hof.update(population)
        record = stats.compile(population)
        logbook.record(gen=g, evals=len(invalid_ind), **record)
        print(logbook.stream)

    best_network = sn.SortingNetwork(INPUTS, hof[0])
    print(stats)
    print(best_network)
    print(best_network.draw())
    print("%i errors, length %i, depth %i" % hof[0].fitness.values)
    
    return population, logbook, hof

if __name__ == "__main__":
    main()


#
#    This file is part of Scalable COncurrent Operations in Python (SCOOP).
#
#    SCOOP is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation, either version 3 of
#    the License, or (at your option) any later version.
#
#    SCOOP is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with SCOOP. If not, see <http://www.gnu.org/licenses/>.
#
"""
Example of shared constants use.
"""
from scoop import futures, shared


def getValue(words):
    """Computes the sum of the values of the words."""
    value = 0
    for word in words:
        for letter in word:
            # shared.getConst will evaluate to the dictionary broadcasted by
            # the root Future
            value += shared.getConst('lettersValue')[letter]
    return value


if __name__ == "__main__":
    # Set the values of the letters according to the language and broadcast it
    # This list is set at runtime
    import sys
    if len(sys.argv) > 1 and sys.argv[1] == 'francais':
        shared.setConst(lettersValue={'a': 1, 'b': 3, 'c': 3, 'd': 2, 'e': 1,
        'f': 4, 'g': 2, 'h': 4, 'i': 1, 'j': 8, 'k':10, 'l': 1, 'm': 2, 'n': 1,
        'o': 1, 'p': 3, 'q': 8, 'r': 1, 'r': 1, 's': 1, 't': 1, 'u': 1, 'v': 4,
        'w':10, 'x':10, 'y':10, 'z': 10})
        print("French letter values used.")
    else:
        shared.setConst(lettersValue={'a': 1, 'b': 3, 'c': 3, 'd': 2, 'e': 1,
        'f': 4, 'g': 2, 'h': 4, 'i': 1, 'j': 8, 'k': 5, 'l': 1, 'm': 3, 'n': 1,
        'o': 1, 'p': 3, 'q':10, 'r': 1, 'r': 1, 's': 1, 't': 1, 'u': 1, 'v': 4,
        'w': 4, 'x': 8, 'y': 4, 'z': 10})
        print("English letter values used.")

    # Get the player words (generate a list of random letters 
    import random
    import string
    random.seed(3141592653)
    words = []
    player_quantity = 4
    words_per_player = 10
    word_letters = (1, 6)
    for pid in range(player_quantity):
        player = []
        for _ in range(words_per_player):
            word = "".join(random.choice(string.ascii_lowercase) for _ in range(random.randint(*word_letters)))
            player.append(word)
        print("Player {pid} played words: {player}".format(**locals()))
        words.append(player)
    
    # Compute the score of every player and display it
    results = list(futures.map(getValue, words))
    for pid, result in enumerate(results):
        print("Player {pid}: {result}".format(**locals()))



#!/usr/bin/env python
#
#    This file is part of Scalable COncurrent Operations in Python (SCOOP).
#
#    SCOOP is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation, either version 3 of
#    the License, or (at your option) any later version.
#
#    SCOOP is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with SCOOP. If not, see <http://www.gnu.org/licenses/>.
#
from collections import deque, defaultdict
import time
import zmq
import sys
import copy
import logging
try:
    import cPickle as pickle
except ImportError:
    import pickle

import scoop
from scoop import TIME_BETWEEN_PARTIALDEBUG
from .. import discovery, utils
from .structs import BrokerInfo


# Worker requests
INIT = b"I"
REQUEST = b"RQ"
TASK = b"T"
REPLY = b"RP"
SHUTDOWN = b"S"
VARIABLE = b"V"
BROKER_INFO = b"B"
STATUS_REQ = b"SR"
STATUS_ANS = b"SA"
STATUS_DONE = b"SD"
STATUS_UPDATE = b"SU"

# Task statuses
STATUS_HERE = b"H"
STATUS_GIVEN = b"G"
STATUS_NONE = b"N"

# Broker interconnection
CONNECT = b"C"


class LaunchingError(Exception): pass


class Broker(object):
    def __init__(self, tSock="tcp://*:*", mSock="tcp://*:*", debug=False,
                 headless=False, hostname="127.0.0.1"):
        """This function initializes a broker.

        :param tSock: Task Socket Address.
        Must contain protocol, address  and port information.
        :param mSock: Meta Socket Address.
        Must contain protocol, address and port information.
        """
        # Initialize zmq
        self.context = zmq.Context(1)

        self.debug = debug
        self.hostname = hostname

        # zmq Socket for the tasks, replies and request.
        self.task_socket = self.context.socket(zmq.ROUTER)
        self.task_socket.setsockopt(zmq.IPV4ONLY, 0)
        self.task_socket.setsockopt(zmq.ROUTER_MANDATORY, 1)
        self.task_socket.setsockopt(zmq.LINGER, 1000)
        self.t_sock_port = 0
        if tSock[-2:] == ":*":
            self.t_sock_port = self.task_socket.bind_to_random_port(tSock[:-2])
        else:
            self.task_socket.bind(tSock)
            self.t_sock_port = tSock.split(":")[-1]


        # Create identifier for this broker
        self.name = "{0}:{1}".format(hostname, self.t_sock_port)

        # Initialize broker logging
        self.logger = utils.initLogging(2 if debug else 0, name=self.name)
        self.logger.handlers[0].setFormatter(
            logging.Formatter(
                "[%(asctime)-15s] %(module)-9s ({0}) %(levelname)-7s "
                "%(message)s".format(self.name)
            )
        )

        # zmq Socket for the pool informations
        self.info_socket = self.context.socket(zmq.PUB)
        self.info_socket.setsockopt(zmq.IPV4ONLY, 0)
        self.info_socket.setsockopt(zmq.LINGER, 1000)
        self.info_sock_port = 0
        if mSock[-2:] == ":*":
            self.info_sock_port = self.info_socket.bind_to_random_port(mSock[:-2])
        else:
            self.info_socket.bind(mSock)
            self.info_sock_port = mSock.split(":")[-1]

        self.task_socket.setsockopt(zmq.SNDHWM, 0)
        self.task_socket.setsockopt(zmq.RCVHWM, 0)
        self.info_socket.setsockopt(zmq.SNDHWM, 0)
        self.info_socket.setsockopt(zmq.RCVHWM, 0)

        # Init connection to fellow brokers
        self.cluster_socket = self.context.socket(zmq.DEALER)
        self.cluster_socket.setsockopt(zmq.IPV4ONLY, 0)
        self.cluster_socket.setsockopt_string(zmq.IDENTITY, self.getName())
            
        self.cluster_socket.setsockopt(zmq.RCVHWM, 0)
        self.cluster_socket.setsockopt(zmq.SNDHWM, 0)
        self.cluster_socket.setsockopt(zmq.IMMEDIATE, 1)

        self.cluster = []
        self.cluster_available = set()

        # Init statistics
        if self.debug:
            self.stats = []
            self.lastDebugTs = time.time()

        # Two cases are important and must be optimised:
        # - The search of unassigned task
        # - The search of available workers
        # These represent when the broker must deal the communications the
        # fastest. Other cases, the broker isn't flooded with urgent messages.

        # Initializing the queue of workers and tasks
        # The busy workers variable will contain a dict (map) of workers: task
        self.available_workers = deque()
        self.unassigned_tasks = deque()
        self.assigned_tasks = defaultdict(set)
        self.status_times = {}
        # Shared variables containing {workerID:{varName:varVal},}
        self.shared_variables = defaultdict(dict)

        # Start a worker-like communication if needed
        self.execQueue = None

        # Handle cloud-like behavior
        self.discovery_thread = None
        self.config = defaultdict(bool)
        self.processConfig({'headless': headless})

    def addBrokerList(self, aBrokerInfoList):
        """Add a broker to the broker cluster available list.
        Connects to the added broker if needed."""
        self.cluster_available.update(set(aBrokerInfoList))

        # If we need another connection to a fellow broker
        # TODO: only connect to a given number
        for aBrokerInfo in aBrokerInfoList:
            self.cluster_socket.connect(
                "tcp://{hostname}:{port}".format(
                    hostname=aBrokerInfo.hostname,
                    port=aBrokerInfo.task_port,
                )
            )
            self.cluster.append(aBrokerInfo)

    def processConfig(self, worker_config):
        """Update the pool configuration with a worker configuration.
        """
        self.config['headless'] |= worker_config.get("headless", False)
        if self.config['headless']:
            # Launch discovery process
            if not self.discovery_thread:
                self.discovery_thread = discovery.Advertise(
                    port=",".join(str(a) for a in self.getPorts()),
                )

    def run(self):
        """Redirects messages until a shutdown message is received."""
        while True:
            if not self.task_socket.poll(-1):
                continue

            msg = self.task_socket.recv_multipart()
            msg_type = msg[1]

            if self.debug:
                self.stats.append((time.time(),
                                   msg_type,
                                   len(self.unassigned_tasks),
                                   len(self.available_workers)))
                if time.time() - self.lastDebugTs > TIME_BETWEEN_PARTIALDEBUG:
                    self.writeDebug("debug/partial-{0}".format(
                        round(time.time(), -1)
                    ))
                    self.lastDebugTs = time.time()

            # New task inbound
            if msg_type == TASK:
                task_id = msg[2]
                task = msg[3]
                self.logger.debug("Received task {0}".format(task_id))
                try:
                    address = self.available_workers.popleft()
                except IndexError:
                    self.unassigned_tasks.append((task_id, task))
                else:
                    self.logger.debug("Sent {0}".format(task_id))
                    self.task_socket.send_multipart([address, TASK, task])
                    self.assigned_tasks[address].add(task_id)

            # Request for task
            elif msg_type == REQUEST:
                address = msg[0]
                try:
                    task_id, task = self.unassigned_tasks.popleft()
                except IndexError:
                    self.available_workers.append(address)
                else:
                    self.logger.debug("Sent {0}".format(task_id))
                    self.task_socket.send_multipart([address, TASK, task])
                    self.assigned_tasks[address].add(task_id)

            # A task status request is requested
            elif msg_type == STATUS_REQ:
                self.pruneAssignedTasks()
                address = msg[0]
                task_id = msg[2]

                if any(task_id in x for x in self.assigned_tasks.values()):
                    status = STATUS_GIVEN
                elif task_id in (x[0] for x in self.unassigned_tasks):
                    status = STATUS_HERE
                else:
                    status = STATUS_NONE

                self.task_socket.send_multipart([
                    address, STATUS_ANS, task_id, status
                ])

            # A task status set (task done) is received
            elif msg_type == STATUS_DONE:
                address = msg[0]
                task_id = msg[2]

                try:
                    self.assigned_tasks[address].discard(task_id)
                except KeyError:
                    pass

            elif msg_type == STATUS_UPDATE:
                address = msg[0]
                try:
                    tasks_ids = pickle.loads(msg[2])
                except:
                    self.logger.error("Could not unpickle status update message.")
                else:
                    self.assigned_tasks[address] = tasks_ids
                    self.status_times[address] = time.time()

            # Answer needing delivery
            elif msg_type == REPLY:
                self.logger.debug("Relaying")
                destination = msg[-1]
                origin = msg[0]
                self.task_socket.send_multipart([destination] + msg[1:] + [origin])

            # Shared variable to distribute
            elif msg_type == VARIABLE:
                address = msg[4]
                value = msg[3]
                key = msg[2]
                self.shared_variables[address].update(
                    {key: value},
                )
                self.info_socket.send_multipart([VARIABLE,
                                                key,
                                                value,
                                                address])

            # Initialize the variables of a new worker
            elif msg_type == INIT:
                address = msg[0]
                try:
                    self.processConfig(pickle.loads(msg[2]))
                except pickle.PickleError:
                    continue
                self.task_socket.send_multipart([
                    address,
                    pickle.dumps(self.config,
                                 pickle.HIGHEST_PROTOCOL),
                    pickle.dumps(self.shared_variables,
                                 pickle.HIGHEST_PROTOCOL),
                ])

                self.task_socket.send_multipart([
                    address,
                    pickle.dumps(self.cluster_available,
                                 pickle.HIGHEST_PROTOCOL),
                ])

            # Add a given broker to its fellow list
            elif msg_type == CONNECT:
                try:
                    connect_brokers = pickle.loads(msg[2])
                except pickle.PickleError:
                    self.logger.error("Could not understand CONNECT message.")
                    continue
                self.logger.info("Connecting to other brokers...")
                self.addBrokerList(connect_brokers)

            # Shutdown of this broker was requested
            elif msg_type == SHUTDOWN:
                self.logger.debug("SHUTDOWN command received.")
                self.shutdown()
                break

    def pruneAssignedTasks(self):
        to_keep = set()
        for address in self.assigned_tasks.keys():
            addr_time = self.status_times.get(address, 0)
            if addr_time + scoop.TIME_BETWEEN_STATUS_PRUNING > time.time():
                to_keep.add(address)

        to_remove = set(self.assigned_tasks.keys()).difference(to_keep)
        for addr in to_remove:
            self.assigned_tasks.pop(addr)

        to_remove = set(self.status_times.keys()).difference(to_keep)
        for addr in to_remove:
            self.status_times.pop(addr)

    def getPorts(self):
        return (self.t_sock_port, self.info_sock_port)

    def getName(self):
        import sys
        if sys.version < '3':
            return unicode(self.name)
        return self.name

    def shutdown(self):
        # This send may raise an ZMQError
        # Looping over it until it gets through
        for i in range(100):
            try:
                self.info_socket.send(SHUTDOWN)
            except zmq.ZMQError:
                time.sleep(0.01)
            else:
                break
        time.sleep(0.1)

        self.context.destroy(1000)

        # Write down statistics about this run if asked
        if self.debug:
            self.writeDebug()

    def writeDebug(self, path="debug"):
        import os
        import pickle
        try:
            os.makedirs(path)
        except:
            pass
        name = self.name.replace(":", "_")
        with open(os.path.join(
                path,
                "broker-{name}".format(**locals())), 'wb') as f:
            pickle.dump(self.stats, f)



#
#    This file is part of Scalable COncurrent Operations in Python (SCOOP).
#
#    SCOOP is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation, either version 3 of
#    the License, or (at your option) any later version.
#
#    SCOOP is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with SCOOP. If not, see <http://www.gnu.org/licenses/>.
#
#    This file is part of DEAP.
#
#    DEAP is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation, either version 3 of
#    the License, or (at your option) any later version.
#
#    DEAP is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with DEAP. If not, see <http://www.gnu.org/licenses/>.
"""
Code from the deap framework, available at:
https://code.google.com/p/deap/source/browse/examples/ga/onemax_short.py
Conversion to its parallel form took two lines:
from scoop import futures
toolbox.register("map", futures.map)
"""
import array
import random

import numpy

from deap import algorithms
from deap import base
from deap import creator
from deap import tools
from scoop import futures

creator.create("FitnessMax", base.Fitness, weights=(1.0,))
creator.create("Individual", array.array, typecode='b', fitness=creator.FitnessMax)

toolbox = base.Toolbox()

# Attribute generator
toolbox.register("attr_bool", random.randint, 0, 1)

# Structure initializers
toolbox.register("individual", tools.initRepeat, creator.Individual, toolbox.attr_bool, 100)
toolbox.register("population", tools.initRepeat, list, toolbox.individual)

def evalOneMax(individual):
    return sum(individual),

toolbox.register("evaluate", evalOneMax)
toolbox.register("mate", tools.cxTwoPoint)
toolbox.register("mutate", tools.mutFlipBit, indpb=0.05)
toolbox.register("select", tools.selTournament, tournsize=3)
toolbox.register("map", futures.map)

def main():
    random.seed(64)
    
    pop = toolbox.population(n=300)
    hof = tools.HallOfFame(1)
    stats = tools.Statistics(lambda ind: ind.fitness.values)
    stats.register("avg", numpy.mean)
    stats.register("std", numpy.std)
    stats.register("min", numpy.min)
    stats.register("max", numpy.max)
    
    pop, log = algorithms.eaSimple(pop, toolbox, cxpb=0.5, mutpb=0.2, ngen=40, 
                                   stats=stats, halloffame=hof, verbose=True)
    
    return pop, log, hof

if __name__ == "__main__":
    main()


#
#    This file is part of Scalable COncurrent Operations in Python (SCOOP).
#
#    SCOOP is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as
#    published by the Free Software Foundation, either version 3 of
#    the License, or (at your option) any later version.
#
#    SCOOP is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public
#    License along with SCOOP. If not, see <http://www.gnu.org/licenses/>.
#
import os
import sys
try:
    import cPickle as pickle
except ImportError:
    import pickle

import scoop


def getDebugIdentifier():
    """Returns the unique identifier of the current worker."""
    return scoop.worker.decode().replace(":", "_")


def getDebugDirectory():
    """Returns the debug directory."""
    return os.path.join(scoop.WORKING_DIRECTORY, "debug")


def createDirectory(path_suffix=""):
    """Create a directory in a way that multiple concurrent requests won't
    be problematic."""
    try:
        os.makedirs(os.path.join(getDebugDirectory(), path_suffix))
    except:
        pass


def redirectSTDOUTtoDebugFile():
    """Redirects the stdout and stderr of the current process to a file."""
    import sys
    kwargs = {}
    if sys.version_info >= (3,):
        kwargs["encoding"] = "utf8"
    sys.stdout = open(
        os.path.join(
            getDebugDirectory(),
            "{0}.stdout".format(getDebugIdentifier()),
        ),
        "w",
        1, # Buffering by line
        **kwargs
    )
    sys.stderr = open(
        os.path.join(
            getDebugDirectory(),
            "{0}.stderr".format(getDebugIdentifier()),
        ),
        "w",
        1, # Buffering by line
        **kwargs
    )


def writeWorkerDebug(debugStats, queueLength, path_suffix=""):
    """Serialize the execution data using pickle and writes it into the debug
    directory."""
    createDirectory(path_suffix)
    origin_prefix = "origin-" if scoop.IS_ORIGIN else ""
    statsFilename = os.path.join(
        getDebugDirectory(),
        path_suffix,
        "{1}worker-{0}-STATS".format(getDebugIdentifier(), origin_prefix)
    )
    lengthFilename = os.path.join(
        getDebugDirectory(),
        path_suffix,
        "{1}worker-{0}-QUEUE".format(getDebugIdentifier(), origin_prefix)
    )
    with open(statsFilename, 'wb') as f:
        pickle.dump(debugStats, f)
    with open(lengthFilename, 'wb') as f:
        pickle.dump(queueLength, f)